from visions.types.boolean import Boolean
from visions.types.categorical import Categorical
from visions.types.complex import Complex
from visions.types.count import Count
from visions.types.date import Date
from visions.types.date_time import DateTime
from visions.types.email_address import EmailAddress
from visions.types.file import File
from visions.types.float import Float
from visions.types.generic import Generic
from visions.types.geometry import Geometry
from visions.types.image import Image
from visions.types.integer import Integer
from visions.types.ip_address import IPAddress
from visions.types.object import Object
from visions.types.ordinal import Ordinal
from visions.types.path import Path
from visions.types.string import String
from visions.types.time import Time
from visions.types.time_delta import TimeDelta
from visions.types.type import VisionsBaseType
from visions.types.url import URL
from visions.types.uuid import UUID

__all__ = [
    "VisionsBaseType",
    "Generic",
    "String",
    "Boolean",
    "Categorical",
    "Complex",
    "Count",
    "Date",
    "DateTime",
    "File",
    "Float",
    "Geometry",
    "Image",
    "Integer",
    "IPAddress",
    "Object",
    "Ordinal",
    "Path",
    "TimeDelta",
    "UUID",
    "URL",
    "Time",
    "EmailAddress",
]



from typing import Dict, Set, Tuple, Type

import networkx as nx
import pytest

from visions import VisionsBaseType


def all_series_included(series_list, series_map):
    """Check that all names are indeed used"""
    used_names = set([name for names in series_map.values() for name in names])
    names = set([series.name for series in series_list])
    if not names == used_names:
        raise ValueError(
            "Not all series are included {unused}".format(unused=names ^ used_names)
        )


def get_contains_cases(
    _test_suite, _series_map: Dict[Type[VisionsBaseType], Set[str]], typeset
):
    """Parametrize contains tests

    Args:
        mapping: mapping from type to a set of series' identifiers

    Returns:
        the args for the generated tests
    """

    # Include children's series in parent
    reversed_topological_edge_sort = list(
        reversed(list(nx.topological_sort(nx.line_graph(typeset.base_graph))))
    )
    for parent, child in reversed_topological_edge_sort:
        _series_map[parent] |= _series_map[child]

    all_series_included(_test_suite, _series_map)

    argsvalues = []
    for item in _test_suite:
        for type, series_list in _series_map.items():
            args = {"id": f"{item.name} x {type}"}

            member = item.name in series_list
            argsvalues.append(pytest.param(item, type, member, **args))

    return {"argnames": "series,type,member", "argvalues": argsvalues}


def contains(series, type, member):
    return (
        member == (series in type),
        f"{series.name} in {type}; expected {member}, got {series in type}",
    )


def get_inference_cases(_test_suite, inferred_series_type_map, typeset):
    argsvalues = []
    for series in _test_suite:
        expected_type = inferred_series_type_map[series.name]
        for test_type in typeset.types:
            args = {
                "id": "{name} x {type} expected {expected}".format(
                    name=series.name,
                    type=test_type,
                    expected=test_type == expected_type,
                )
            }
            difference = test_type != expected_type
            argsvalues.append(
                pytest.param(series, test_type, typeset, difference, **args)
            )
    return {"argnames": "series,type,typeset,difference", "argvalues": argsvalues}


def infers(series, expected_type, typeset, difference):
    inferred_type = typeset.infer_type(series)
    return (
        (inferred_type == expected_type) != difference,
        f"inference of {series.name} expected {expected_type} to be {difference} (typeset={typeset})",
    )
    # return series in inferred_type, f"series should be member of inferred type"


def all_relations_tested(series_map, typeset):
    # Convert data structure for mapping
    series_map_lookup = {}
    for map_to_type, map_from_type, items in series_map:
        try:
            series_map_lookup[map_to_type][map_from_type] = items
        except KeyError:
            series_map_lookup[map_to_type] = {map_from_type: items}

    missing_relations = set()
    for node in typeset.types:
        for relation in node.relations:
            from_type, to_type = relation.related_type, relation.type
            if relation.inferential and (
                to_type not in series_map_lookup
                or from_type not in series_map_lookup[to_type]
                or len(series_map_lookup[to_type][from_type]) == 0
            ):
                missing_relations.add(str(relation))

    if len(missing_relations) > 0:
        raise ValueError(
            "Not all inferential relations are tested {missing_relations}".format(
                missing_relations=missing_relations
            )
        )


def get_convert_cases(_test_suite, _series_map, typeset):
    all_relations_tested(_series_map, typeset)

    argsvalues = []
    for item in _test_suite:
        for source_type, relation_type, series_list in _series_map:
            if item in relation_type:
                args = {
                    "id": "{name}: {relation_type} -> {source_type}".format(
                        name=item.name,
                        relation_type=relation_type,
                        source_type=source_type,
                    )
                }
                member = item.name in series_list
                argsvalues.append(
                    pytest.param(source_type, relation_type, item, member, **args)
                )

    return dict(
        argnames=["source_type", "relation_type", "series", "member"],
        argvalues=argsvalues,
    )


def convert(source_type, relation_type, series, member) -> Tuple[bool, str]:
    relation_gen = (
        rel for rel in source_type.relations if rel.related_type == relation_type
    )
    relation = next(relation_gen)

    is_relation = relation.is_relation(series)

    if not member:
        return (
            (not is_relation),
            f"{source_type}, {relation}, {member}, {series.name}, {series[0]}",
        )
    else:
        cast_series = relation.transform(series)

        return (
            (is_relation and cast_series in source_type),
            "Relationship {relation} cast {series_values} to {cast_values}".format(
                relation=relation,
                series_values=series.values,
                cast_values=cast_series.values,
            ),
        )



import pathlib
from typing import Sequence

import pandas as pd

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.series_utils import nullable_series_contains


def string_is_path(series) -> bool:
    try:
        s = to_path(series.copy())
        return s.apply(lambda x: x.is_absolute()).all()
    except TypeError:
        return False


def to_path(series: pd.Series) -> pd.Series:
    s = series.copy().apply(pathlib.PureWindowsPath)
    if not s.apply(lambda x: x.is_absolute()).all():
        return series.apply(pathlib.PurePosixPath)
    else:
        return s


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Object, String

    relations = [
        IdentityRelation(cls, Object),
        InferenceRelation(
            cls, String, relationship=string_is_path, transformer=to_path
        ),
    ]
    return relations


class Path(VisionsBaseType):
    """**Path** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> import pathlib
        >>> x = pd.Series([pathlib.Path('/home/user/file.txt'), pathlib.Path('/home/user/test2.txt')])
        >>> x in visions.Path
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    @nullable_series_contains
    def contains_op(cls, series: pd.Series) -> bool:
        return all(isinstance(x, pathlib.PurePath) and x.is_absolute() for x in series)



import datetime
from urllib.parse import urlparse

import numpy as np
import pandas as pd
import pytest

from visions.application.summaries import CompleteSummary
from visions.types import (
    URL,
    Boolean,
    Categorical,
    Complex,
    DateTime,
    Float,
    Geometry,
    Integer,
    Object,
    String,
)


@pytest.fixture(scope="class")
def summary():
    return CompleteSummary()


def validate_summary_output(test_series, visions_type, correct_output, summary):
    trial_output = summary.summarize_series(test_series, visions_type)

    for metric, result in correct_output.items():
        assert metric in trial_output, "Metric `{metric}` is missing".format(
            metric=metric
        )

        if isinstance(trial_output[metric], pd.Series):
            trial_output[metric] = trial_output[metric].to_dict()

        assert (
            trial_output[metric] == result
        ), "Expected value {result} for metric `{metric}`, got {output}".format(
            result=result, metric=metric, output=trial_output[metric]
        )


def test_integer_summary(summary, visions_type=Integer):
    test_series = pd.Series([0, 1, 2, 3, 4])
    correct_output = {
        "n_unique": 5,
        "mean": 2,
        "median": 2,
        "std": pytest.approx(1.58113, 0.00001),
        "max": 4,
        "min": 0,
        "n_records": 5,
        "n_zeros": 1,
    }

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_integer_missing_summary(summary, visions_type=Integer):
    test_series = pd.Series([0, 1, 2, 3, 4])
    correct_output = {
        "n_unique": 5,
        "mean": 2,
        "median": 2,
        "std": pytest.approx(1.58113, 0.00001),
        "max": 4,
        "min": 0,
        "n_records": 5,
        "n_zeros": 1,
        "na_count": 0,
    }

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_float_missing_summary(summary, visions_type=Float):
    test_series = pd.Series([0.0, 1.0, 2.0, 3.0, 4.0, np.nan])
    correct_output = {
        "n_unique": 5,
        "median": 2,
        "mean": 2,
        "std": pytest.approx(1.58113, 0.00001),
        "max": 4,
        "min": 0,
        "n_records": 6,
        "n_zeros": 1,
        "na_count": 1,
    }

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_bool_missing_summary(summary, visions_type=Boolean):
    test_series = pd.Series([True, False, True, True, np.nan])
    correct_output = {"n_records": 5, "na_count": 1}

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_categorical_missing_summary(summary, visions_type=Categorical):
    test_series = pd.Series(
        pd.Categorical(
            [True, False, np.nan, "test"],
            categories=[True, False, "test", "missing"],
            ordered=True,
        )
    )
    correct_output = {
        "n_unique": 3,
        "n_records": 4,
        "na_count": 1,
        "category_size": 4,
        "missing_categorical_values": True,
    }

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_complex_missing_summary(summary, visions_type=Complex):
    test_series = pd.Series([0 + 0j, 0 + 1j, 1 + 0j, 1 + 1j, np.nan])
    correct_output = {"n_unique": 4, "mean": 0.5 + 0.5j, "na_count": 1, "n_records": 5}

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_datetime_missing_summary(summary, visions_type=DateTime):
    test_series = pd.Series(
        [
            datetime.datetime(2010, 1, 1),
            datetime.datetime(2010, 8, 2),
            datetime.datetime(2011, 2, 1),
            np.nan,
        ]
    )
    correct_output = {
        "n_unique": 3,
        "max": datetime.datetime(2011, 2, 1),
        "min": datetime.datetime(2010, 1, 1),
        "n_records": 4,
        "na_count": 1,
        "range": test_series.max() - test_series.min(),
    }

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_object_missing_summary(summary, visions_type=Object):
    test_series = pd.Series([datetime.datetime(2010, 1, 1), "test", 3, np.nan])
    correct_output = {
        "n_unique": 3,
        "frequencies": {"test": 1, 3: 1, datetime.datetime(2010, 1, 1): 1},
        "n_records": 4,
        "na_count": 1,
    }

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_geometry_missing_summary(summary, visions_type=Geometry):
    test_series = pd.Series([np.nan])
    correct_output = {"na_count": 1, "n_records": 1}

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_string_missing_summary(summary, visions_type=String):
    test_series = pd.Series(["apple", "orange", "bike", np.nan])
    correct_output = {"na_count": 1, "n_unique": 3, "n_records": 4}

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_string_summary(summary, visions_type=String):
    test_series = pd.Series(["http://ru.nl", "http://ru.nl", "http://nl.ru"])
    correct_output = {"n_unique": 2, "n_records": 3}

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_string_empty_summary(summary, visions_type=String):
    test_series = pd.Series(["", "", ""])
    correct_output = {"n_unique": 1, "n_records": 3}

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_url_summary(summary, visions_type=URL):
    test_series = pd.Series(
        [urlparse("http://ru.nl"), urlparse("http://ru.nl"), urlparse("http://nl.ru")]
    )
    correct_output = {
        "n_records": 3,
        "n_unique": 2,
        "scheme_counts": {"http": 3},
        "netloc_counts": {"ru.nl": 2, "nl.ru": 1},
        "path_counts": {"": 3},
        "query_counts": {"": 3},
        "fragment_counts": {"": 3},
    }

    validate_summary_output(test_series, visions_type, correct_output, summary)


def test_empty_summary():
    pass



from typing import Sequence

import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.coercion import test_utils


def to_datetime(series: pd.Series) -> pd.Series:
    return pd.to_datetime(series)


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Generic, String

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(
            cls,
            String,
            relationship=test_utils.coercion_test(
                to_datetime, [OverflowError, TypeError]  # type: ignore
            ),
            transformer=to_datetime,
        ),
    ]
    return relations


class DateTime(VisionsBaseType):
    """**Datetime** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([pd.datetime(2017, 3, 5), pd.datetime(2019, 12, 4)])
        >>> x in visions.DateTime
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return pdt.is_datetime64_any_dtype(series)



from pprint import pprint

import numpy as np
import pandas as pd

import visions as v
from visions.application.summaries import CompleteSummary

category_series = pd.Series(
    pd.Categorical(
        [True, False, np.nan, "test"], categories=[True, False, "test", "missing"]
    )
)

# Generate a summary
summarizer = CompleteSummary()
summary = summarizer.summarize_series(category_series, v.Categorical)

pprint(summary)



import pytest

from visions import Float, Generic, VisionsBaseType, VisionsTypeset
from visions.relations import IdentityRelation


class CustomGeneric(Generic):
    this_value = True


class CustomNonGeneric:
    another_value = False

    @classmethod
    def get_relations(cls):
        return []


def make_test_type(root):
    class CustomFloat(VisionsBaseType):
        @classmethod
        def contains_op(cls, series):
            return True

        @classmethod
        def get_relations(cls):
            return [IdentityRelation(cls, root)]

    return CustomFloat


def test_root_node():
    class CustomSet(VisionsTypeset):
        def __init__(self):
            super().__init__({make_test_type(CustomGeneric), CustomGeneric})

    _ = CustomSet()


def test_root_node_other():
    class CustomSet(VisionsTypeset):
        def __init__(self):
            super().__init__({make_test_type(CustomNonGeneric), CustomNonGeneric})

    with pytest.raises(ValueError, match="`root_node` should be a subclass of Generic"):
        _ = CustomSet()


def test_multiple_roots():
    class CustomSet(VisionsTypeset):
        def __init__(self):
            super().__init__({make_test_type(CustomGeneric), CustomGeneric, Generic})

    _ = CustomSet()



from pathlib import Path

from visions.application.summaries.complete_summary import CompleteSummary
from visions.types import DateTime, File, Integer, String


def generate_summary_plots():
    summaries_dir = Path("summaries/")
    summaries_dir.mkdir(exist_ok=True)

    summary = CompleteSummary()

    # For a typeset
    summary.plot(summaries_dir / "summary_complete.svg")

    # For specific types
    summary.plot(summaries_dir / "summary_integer.svg", Integer)
    summary.plot(summaries_dir / "summary_string.svg", String)
    summary.plot(summaries_dir / "summary_datetime.svg", DateTime)
    summary.plot(summaries_dir / "summary_file.svg", File)


if __name__ == "__main__":
    generate_summary_plots()



import warnings
from functools import singledispatch
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Type, TypeVar, Union

import networkx as nx
import pandas as pd

from visions.types.generic import Generic
from visions.types.type import VisionsBaseType

pdT = TypeVar("pdT", pd.Series, pd.DataFrame)
TypeOrTypeset = TypeVar("TypeOrTypeset", Type[VisionsBaseType], "VisionsTypeset")
pathTypes = TypeVar(
    "pathTypes", Type[VisionsBaseType], Dict[str, Type[VisionsBaseType]]
)


def build_graph(nodes: Set[Type[VisionsBaseType]]) -> Tuple[nx.DiGraph, nx.DiGraph]:
    """Constructs a traversable relation graph between visions types

    Builds a type relation graph from a collection of :class:`visions.types.type.VisionsBaseType` where
    each node corresponds to a type and each edge is a relation defined on the type.

    Args:
        nodes:  An iterable of :class:`visions.types.type.VisionsBaseType`

    Returns:
        A directed graph of type relations for the provided nodes.
    """

    style_map = {True: "dashed", False: "solid"}
    relation_graph = nx.DiGraph()
    relation_graph.add_nodes_from(nodes)

    noninferential_edges = []

    for node in nodes:
        for relation in node.get_relations():
            if relation.related_type not in nodes:
                warnings.warn(
                    "Provided relations included mapping from {related_type} to {own_type} but {related_type} was not included in the provided list of nodes".format(
                        related_type=relation.related_type, own_type=relation.type
                    )
                )
            else:
                relation_graph.add_edge(
                    relation.related_type,
                    relation.type,
                    relationship=relation,
                    style=style_map[relation.inferential],
                )

                if not relation.inferential:
                    noninferential_edges.append((relation.related_type, relation.type))

    check_graph_constraints(relation_graph)

    base_graph = relation_graph.edge_subgraph(noninferential_edges)
    return relation_graph, base_graph


def check_graph_constraints(relation_graph: nx.DiGraph) -> None:
    """Validates a relation_graph is appropriately constructed

    Args:
        relation_graph: A directed graph representing the set of relations between type nodes.

    """
    check_isolates(relation_graph)
    check_cycles(relation_graph)


def check_isolates(graph: nx.DiGraph) -> None:
    """Check for orphaned nodes.

    Args:
        graph: the graph to check

    """
    nodes = set(graph.nodes)
    root_node = next(nx.topological_sort(graph))

    isolates = list(set(nx.isolates(graph)) - {root_node})  # root can be isolate
    graph.remove_nodes_from(isolates)
    orphaned_nodes = nodes - set(graph.nodes)
    if orphaned_nodes:
        message = f"{orphaned_nodes} were isolates in the type relation map and consequently orphaned. "
        message += "Please add some mapping to the orphaned nodes."
        warnings.warn(message)


def check_cycles(graph: nx.DiGraph) -> None:
    """Check for cycles and warn if one is found

    Args:
        graph: the graph to check

    """
    cycles = list(nx.simple_cycles(graph))
    if len(cycles) > 0:
        warnings.warn(
            "Cyclical relations between types {cycles} detected".format(cycles=cycles)
        )


def traverse_graph_with_series(
    base_type: Type[VisionsBaseType],
    series: pd.Series,
    graph: nx.DiGraph,
    path: List[Type[VisionsBaseType]] = None,
) -> Tuple[pd.Series, List[Type[VisionsBaseType]]]:
    """Depth First Search traversal. There should be at most one successor that contains the series.

    Args:
        base_type: Entry-point for graph to start traversal
        series: the Series to check
        graph: the Graph to traverse
        path: the path so far

    Returns:
        The most uniquely specified node matching the series.
    """
    if path is None:
        path = []

    path.append(base_type)

    for vision_type in graph.successors(base_type):
        relation = graph[base_type][vision_type]["relationship"]
        if relation.is_relation(series):
            series = relation.transform(series)
            return traverse_graph_with_series(vision_type, series, graph, path)

    return series, path


def traverse_graph_with_sampled_series(
    base_type: Type[VisionsBaseType],
    series: pd.Series,
    graph: nx.DiGraph,
    sample_size: int = 10,
) -> Tuple[pd.Series, List[Type[VisionsBaseType]]]:
    """Depth First Search traversal with sampling. There should be at most one successor that contains the series.

    Args:
        base_type: Entry-point for graph to start  traversal
        series: the Series to check
        graph: the Graph to traverse
        sample_size: number of items used in heuristic traversal

    Returns:
        The most uniquely specified node matching the series.
    """

    if (series.shape[0] < 1000) or (sample_size > series.shape[0]):
        return traverse_graph_with_series(base_type, series, graph)

    series_sample = series.sample(sample_size)
    _, path = traverse_graph_with_series(base_type, series_sample, graph)
    if len(path) == 1:
        return series, path

    # Cast the full series
    from_type = path[0]
    for i, to_type in enumerate(path[1:]):
        relation = graph[from_type][to_type]["relationship"]
        if not relation.is_relation(series):
            break
        series = relation.transform(series)
        from_type = to_type

    return series, path[0 : (i + 2)]


@singledispatch
def traverse_graph(
    data: pdT, root_node: Type[VisionsBaseType], graph: nx.DiGraph
) -> Tuple[pdT, Any]:
    raise TypeError(f"Undefined graph traversal over data of type {type(data)}")


@traverse_graph.register(pd.Series)
def _(
    series: pd.Series, root_node: Type[VisionsBaseType], graph: nx.DiGraph
) -> Tuple[pd.Series, List[Type[VisionsBaseType]]]:
    return traverse_graph_with_series(root_node, series, graph)


@traverse_graph.register(pd.DataFrame)  # type: ignore
def _(
    df: pd.DataFrame, root_node: Type[VisionsBaseType], graph: nx.DiGraph
) -> Tuple[pd.DataFrame, Dict[str, List[Type[VisionsBaseType]]]]:
    inferred_values = {
        col: traverse_graph(df[col], root_node, graph) for col in df.columns
    }
    inferred_paths = {col: inf_path for col, (_, inf_path) in inferred_values.items()}
    inferred_series = {
        col: inf_series for col, (inf_series, _) in inferred_values.items()
    }
    return pd.DataFrame(inferred_series), inferred_paths


@singledispatch
def get_type_from_path(path_data: Any) -> pathTypes:
    raise TypeError(f"Can't get types from path object of type {type(path_data)}")


@get_type_from_path.register(dict)  # type: ignore
def _(path_dict: dict) -> Dict[str, Type[VisionsBaseType]]:
    return {k: get_type_from_path(v) for k, v in path_dict.items()}


@get_type_from_path.register(list)  # type: ignore
def _(path_list: list) -> Type[VisionsBaseType]:
    return path_list[-1]


@get_type_from_path.register(tuple)  # type: ignore
def _(path_list: tuple) -> Type[VisionsBaseType]:
    return path_list[-1]


class VisionsTypeset(object):
    """
    A collection of :class:`visions.types.type.VisionsBaseType` with  associated relationship map between them.

    Attributes:
        types: The collection of Visions Types derived from :class:`visions.types.type.VisionsBaseType`
        base_graph: The graph of relations composed exclusively of :class:`visions.relations.relations.IdentityRelation`
        relation_graph: The full relation graph including both :class:`visions.relations.relations.IdentityRelation` and :class:`visions.relations.relations.InferenceRelation`
    """

    def __init__(self, types: Set[Type[VisionsBaseType]]) -> None:
        """
        Args:
            types: a set of types
        """
        self._root_node = None

        if not isinstance(types, Iterable):
            raise ValueError("types should be iterable")

        self.relation_graph, self.base_graph = build_graph(set(types))

        if not issubclass(self.root_node, Generic):
            raise ValueError("`root_node` should be a subclass of Generic")

        self.types = set(self.relation_graph.nodes)

    @property
    def root_node(self) -> Type[VisionsBaseType]:
        """Returns a cached copy of the relation_graphs root node
        
        Args:

        Returns:
            A cached copy of the relation_graphs root node.
        """
        if self._root_node is None:
            self._root_node = next(nx.topological_sort(self.relation_graph))
        return self._root_node  # type: ignore

    @staticmethod
    def _traverse_graph(data: pdT, root_node, graph):
        return traverse_graph(data, root_node, graph)

    def detect_type(self, data: pdT) -> pathTypes:
        """The inferred type found only considering IdentityRelations.

        Args:
            data: a DataFrame or Series to determine types over

        Returns:
            A dictionary of {name: type} pairs in the case of DataFrame input or a type
        """
        _, paths = self._traverse_graph(data, self.root_node, self.base_graph)
        return get_type_from_path(paths)

    def infer_type(self, data: pdT) -> pathTypes:
        """The inferred type found using all type relations.

        Args:
            data: a DataFrame or Series to determine types over

        Returns:
            A dictionary of {name: type} pairs in the case of DataFrame input or a type
        """
        _, paths = self._traverse_graph(data, self.root_node, self.relation_graph)
        return get_type_from_path(paths)

    def cast_to_detected(self, data: pdT) -> pdT:
        """Transforms input data into a canonical representation using only IdentityRelations

        Args:
            data: a DataFrame or Series to determine types over

        Returns:
            new_data: The transformed DataFrame or Series.
        """
        data, _ = self._traverse_graph(data, self.root_node, self.base_graph)
        return data

    def cast_to_inferred(self, data: pdT) -> Tuple[pdT, pathTypes]:
        """Transforms input data and returns it's corresponding new type relation using all relations.

        Args:
            data: a DataFrame or Series to determine types over

        Returns:
            new_data: The transformed DataFrame or Series.
            types: A dictionary of {name: type} pairs in the case of DataFrame input or a type.
        """
        data, _ = self._traverse_graph(data, self.root_node, self.relation_graph)
        return data

    def output_graph(
        self,
        file_name: Union[str, Path],
        base_only: bool = False,
        dpi: Optional[int] = None,
    ) -> None:
        """Write the type graph to a file.

        Args:
            file_name: the file to save the output to
            base_only: if True, plot the graph without relation mapping edges
            dpi: set the dpi of the output image
        """
        from visions.utils.graph import output_graph

        if base_only:
            graph = self.base_graph.copy()
        else:
            graph = self.relation_graph.copy()

        graph.graph["node"] = {"shape": "box", "color": "red"}
        if dpi is not None:
            graph.graph["graph"] = {"dpi": dpi}

        output_graph(graph, file_name)

    def plot_graph(self, dpi: int = 800, base_only: bool = False):
        """

        Args:
            dpi: dpi of the matplotlib figure.
            base_only: Only display the typesets base_graph
        Returns:
            Displays the image
        """
        import os
        import tempfile

        from matplotlib import image as mpimg
        from matplotlib import pyplot as plt

        with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as temp_file:
            self.output_graph(temp_file.name, dpi=dpi, base_only=base_only)
            img = mpimg.imread(temp_file.name)
            plt.figure(dpi=dpi)
            plt.axis("off")
            plt.imshow(img)
        os.unlink(temp_file.name)

    def _get_other_type(self, other: TypeOrTypeset) -> Set[Type[VisionsBaseType]]:
        """Converts input into a set of :class:`visions.types.type.VisionsBaseType`

        Args:
            other: A :class:`visions.types.type.VisionsBaseType` or :class:`visions.typesets.typeset.VisionsTypeset`

        Raises:
            NotImplementedError:

        Returns:
            Set[Type[VisionsBaseType]]:
        """
        if isinstance(other, VisionsTypeset):
            other_types = set(other.types)
        elif issubclass(other, VisionsBaseType):
            other_types = {other}
        else:
            raise NotImplementedError(
                "Typeset operation not implemented for type {other_type}".format(
                    other_type=type(other)
                )
            )
        return other_types

    def replace(self, old: VisionsBaseType, new: VisionsBaseType) -> "VisionsTypeset":
        """Create a new typeset having replace one type with another.

        Args:
            old: Visions type to replace.
            new: Replacement visions type.
        
        Returns
            A VisionsTypeset
        """
        types = self.types.copy()
        types.add(new)
        types.remove(old)
        return VisionsTypeset(types)

    def __add__(self, other: TypeOrTypeset) -> "VisionsTypeset":
        """Adds a type or typeset into the current typeset.

        Args:
            other: Type or typeset to be added
        
        Returns
            A VisionsTypeset
        """
        other_types = self._get_other_type(other)
        return VisionsTypeset(self.types | other_types)

    def __iadd__(self, other: TypeOrTypeset) -> "VisionsTypeset":
        """Adds a type or typeset into the current typeset.

        Args:
            other: Type or typeset to be added
        
        Returns
            A VisionsTypeset
        """
        return self.__add__(other)

    def __sub__(self, other: TypeOrTypeset) -> "VisionsTypeset":
        """Subtracts a type or typeset from the current typeset.

        Args:
            other: Type or typeset to be removed
        
        Returns
            A VisionsTypeset
        """
        other_types = self._get_other_type(other)
        return VisionsTypeset(self.types - other_types)

    def __isub__(self, other: TypeOrTypeset) -> "VisionsTypeset":
        """Subtracts a type or typeset from the current typeset.

        Args:
            other: Type or typeset to be removed
        
        Returns
            A VisionsTypeset
        """
        return self.__sub__(other)

    def __repr__(self) -> str:
        """Pretty representation of the typeset.
        
        Returns
            A :class:`visions.typesets.typeset.VisionsTypeset`
        """
        return self.__class__.__name__



from typing import Type

import pandas as pd

from visions.application.summaries.summary import Summary
from visions.types import VisionsBaseType


def summarize_frame(df: pd.DataFrame, series_types: dict, summarizer: Summary):
    return summarizer.summarize_frame(df, {}, series_types)


def summarize_series(
    series: pd.Series, series_type: Type[VisionsBaseType], summarizer: Summary
):
    return summarizer.summarize_series(series, series_type)


def summarize(df: pd.DataFrame, series_types: dict, summarizer: Summary):
    return summarizer.summarize(df, series_types)



from datetime import datetime

import pandas as pd


def file_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """

    stats = series.map(lambda x: x.stat())

    def convert_datetime(x):
        return datetime.fromtimestamp(x).strftime("%Y-%m-%d %H:%M:%S")

    summary = {
        "file_size": stats.map(lambda x: x.st_size),
        "file_created_time": stats.map(lambda x: x.st_ctime).map(convert_datetime),
        "file_accessed_time": stats.map(lambda x: x.st_atime).map(convert_datetime),
        "file_modified_time": stats.map(lambda x: x.st_mtime).map(convert_datetime),
    }
    return summary


def file_hash_summary(series: pd.Series, hash_algorithm="md5") -> dict:
    """Include hash to summary

    Args:
        series: series to summarize
        hash_algorithm: algorithm hashlib uses to calculate the hash

    Returns:
        Summary dict
    """

    import hashlib

    def hash_file(file_name, hash_algorithm):
        hash = hashlib.new(hash_algorithm)
        chunk_size = 64000
        try:
            with open(file_name, "rb") as file:
                file_buffer = file.read(chunk_size)
                while file_buffer:
                    hash.update(file_buffer)
                    file_buffer = file.read(chunk_size)
                return hash.hexdigest()
        except IOError:
            raise

    summary = {
        "hashes_{}".format(hash_algorithm): series.map(
            lambda x: hash_file(x, hash_algorithm)
        )
    }
    return summary



from visions import Integer, IPAddress, VisionsBaseType
from visions.relations import TypeRelation


def test_property_base():
    """Assert that _relations is initialized"""
    assert VisionsBaseType._relations is None


def test_property_types():
    """Assert that the relations property contains a list"""
    assert type(Integer.relations) == list and all(
        isinstance(r, TypeRelation) for r in Integer.relations
    )
    assert type(IPAddress.relations) == list and all(
        isinstance(r, TypeRelation) for r in IPAddress.relations
    )



import pathlib
from typing import Sequence

import pandas as pd

from visions.relations import IdentityRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.series_utils import nullable_series_contains


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Path

    relations = [IdentityRelation(cls, Path)]
    return relations


class File(VisionsBaseType):
    """**File** implementation of :class:`visions.types.type.VisionsBaseType`.
    (i.e. existing path)

    Examples:
        >>> x = pd.Series([pathlib.Path('/home/user/file.txt'), pathlib.Path('/home/user/test2.txt')])
        >>> x in visions.File
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    @nullable_series_contains
    def contains_op(cls, series: pd.Series) -> bool:
        return all(isinstance(p, pathlib.Path) and p.exists() for p in series)



import numpy as np
import pandas as pd

from visions.application.summaries import CompleteSummary
from visions.application.summaries.functional import (
    summarize,
    summarize_frame,
    summarize_series,
)
from visions.types import Integer, String


def test_summarize_frame():
    df = pd.DataFrame(
        {
            "Brand": ["Honda Civic", "Toyota Corolla", "Ford Focus", "Audi A4"],
            "Price": [22000, 25000, 27000, 35000],
        },
        columns=["Brand", "Price"],
    )
    summary = summarize_frame(
        df, {"Brand": String, "Price": Integer}, CompleteSummary()
    )
    summary.pop("type_counts")
    assert summary == {
        "n_observations": 4,
        "n_variables": 2,
        "memory_size": 430,
        "na_count": 0,
        "n_vars_missing": 0,
    }


def test_summarize_series():
    brand_series = pd.Series(["Honda Civic", "Toyota Corolla", "Ford Focus", "Audi A4"])

    summary = summarize_series(brand_series, String, CompleteSummary())

    assert summary["n_unique"] == 4
    assert summary["length"].to_dict() == {0: 11, 1: 14, 2: 10, 3: 7}
    assert summary["max_length"] == 14
    assert summary["min_length"] == 7
    assert summary["mean_length"] == 10.5
    assert summary["median_length"] == 10.5
    assert summary["n_characters"] == 20

    assert summary["category_alias_values"] == {
        "H": "Uppercase_Letter",
        "o": "Lowercase_Letter",
        "n": "Lowercase_Letter",
        "d": "Lowercase_Letter",
        "a": "Lowercase_Letter",
        " ": "Space_Separator",
        "C": "Uppercase_Letter",
        "i": "Lowercase_Letter",
        "v": "Lowercase_Letter",
        "c": "Lowercase_Letter",
        "T": "Uppercase_Letter",
        "y": "Lowercase_Letter",
        "t": "Lowercase_Letter",
        "r": "Lowercase_Letter",
        "l": "Lowercase_Letter",
        "F": "Uppercase_Letter",
        "u": "Lowercase_Letter",
        "s": "Lowercase_Letter",
        "A": "Uppercase_Letter",
        "4": "Decimal_Number",
    }

    assert summary["frequencies"] == {
        "Audi A4": 1,
        "Honda Civic": 1,
        "Ford Focus": 1,
        "Toyota Corolla": 1,
    }

    price_series = pd.Series([22000, 25000, 27000, 35000])

    summary = summarize_series(price_series, Integer, CompleteSummary())
    assert summary == {
        "n_infinite": 0,
        "mean": 27250.0,
        "std": 5560.275772537426,
        "variance": 30916666.666666668,
        "max": 35000.0,
        "min": 22000.0,
        "median": 26000.0,
        "kurt": 1.8192544372679649,
        "skew": 1.19978923754086,
        "sum": 109000.0,
        "mad": 2500.0,
        "quantile_5": 22450.0,
        "quantile_25": 24250.0,
        "quantile_50": 26000.0,
        "quantile_75": 29000.0,
        "quantile_95": 33800.0,
        "iqr": 4750.0,
        "range": 13000.0,
        "cv": 0.20404681734082297,
        "n_zeros": 0,
        "is_unique": True,
        "n_unique": 4,
        "frequencies": {27000: 1, 35000: 1, 25000: 1, 22000: 1},
        "n_records": 4,
        "memory_size": 160,
        "dtype": np.int64,
        "types": {"int": 4},
        "na_count": 0,
        "monotonic_decrease": False,
        "monotonic_decrease_strict": False,
        "monotonic_increase": True,
        "monotonic_increase_strict": True,
    }


def test_summarize():
    df = pd.DataFrame(
        {
            "Brand": ["Honda Civic", "Toyota Corolla", "Ford Focus", "Audi A4"],
            "Price": [22000, 25000, 27000, 35000],
        },
        columns=["Brand", "Price"],
    )
    summary = summarize(df, {"Brand": String, "Price": Integer}, CompleteSummary())

    assert summary["types"] == {"Brand": String, "Price": Integer}



import numpy as np
import pandas as pd

from visions.functional import cast_to_inferred, detect_type, infer_type
from visions.typesets import CompleteSet

# Create a DataFrame from our data
df = pd.DataFrame(
    {
        "numbers_with_nan": [3, 7, np.nan],
        "url": [
            "http://www.cwi.nl:80/%7Eguido/Python.html",
            "https://numpy.org/",
            "https://github.com/pandas-profiling/pandas-profiling",
        ],
        "uuid": [
            "0b8a22ca-80ad-4df5-85ac-fa49c44b7ede",
            "aaa381d6-8442-4f63-88c8-7c900e9a23c6",
            "00000000-0000-0000-0000-000000000000",
        ],
    }
)


# Choose the complete typeset, which includes URLs
typeset = CompleteSet()


# Detect the type (without casting)
print(detect_type(df, typeset))
# {'numbers_with_nan': Float, 'url': String, 'uuid': String}

# Cast the dataframe to inferred types
cast_df = cast_to_inferred(df, typeset)
print(cast_df.to_string())
#    numbers_with_nan                                                url                                  uuid
# 0                 3  (http, www.cwi.nl:80, /%7Eguido/Python.html, ,...  0b8a22ca-80ad-4df5-85ac-fa49c44b7ede
# 1                 7                        (https, numpy.org, /, , , )  aaa381d6-8442-4f63-88c8-7c900e9a23c6
# 2               NaN  (https, github.com, /pandas-profiling/pandas-p...  00000000-0000-0000-0000-000000000000

# Print the inferred types
print(infer_type(df, typeset))
# {'numbers_with_nan': Integer, 'url': URL, 'uuid': UUID}



import uuid
from typing import Sequence

import pandas as pd

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.series_utils import isinstance_attrs, nullable_series_contains


def test_uuid(series) -> bool:
    try:
        return to_uuid(series).all()
    except (ValueError, AttributeError):
        return False


def to_uuid(series: pd.Series) -> pd.Series:
    return series.apply(uuid.UUID)


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Object, String

    relations = [
        IdentityRelation(cls, Object),
        InferenceRelation(cls, String, relationship=test_uuid, transformer=to_uuid),
    ]
    return relations


class UUID(VisionsBaseType):
    """**UUID** implementation of :class:`visions.types.type.VisionsBaseType`.

    References:
        UUID specification in RFC4122:
        https://tools.ietf.org/html/rfc4122#section-3

        Python standard library:
        https://docs.python.org/3/library/uuid.html

    Examples:
        >>> import pandas as pd
        >>> import uuid
        >>> uuids = ['0b8a22ca-80ad-4df5-85ac-fa49c44b7ede', 'aaa381d6-8442-4f63-88c8-7c900e9a23c6']
        >>> x = pd.Series([uuid.UUID(uuid_str) for uuid_str in uuids])
        >>> x in visions.UUID
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    @nullable_series_contains
    def contains_op(cls, series: pd.Series) -> bool:
        return isinstance_attrs(series, uuid.UUID, ["time_low", "hex"])



from visions.relations import InferenceRelation
from visions.types.ordinal import to_ordinal


def check_consecutive(l) -> bool:
    return sorted(l) == list(range(min(l), max(l) + 1))


def is_ordinal_int(s) -> bool:
    unique_values = list(s.unique())
    return (
        check_consecutive(unique_values)
        and 2 < len(unique_values) < 10
        and 1 in unique_values
    )


def integer_to_ordinal(cls) -> InferenceRelation:
    from visions.types import Integer

    return InferenceRelation(
        cls, Integer, relationship=is_ordinal_int, transformer=to_ordinal
    )



# TODO: percentages
# TODO: rename to aggregate?


def dataframe_series_summary(series_summaries: dict) -> dict:
    """Aggregate certain statistics from the series summary

    Args:
        series_summaries: mapping from series to summary

    Returns:
        A aggregated summary based on the statistics of individual series
    """
    summary = {"na_count": 0, "n_vars_missing": 0}
    for series_summary in series_summaries.values():
        if "na_count" in series_summary and series_summary["na_count"] > 0:
            summary["na_count"] += series_summary["na_count"]
    return summary



import imghdr
from pathlib import Path
from typing import Optional, Tuple, Union

import imagehash
from PIL import ExifTags, Image

from visions.utils.monkeypatches.imghdr_patch import *


def open_image(path: Path) -> Optional[Image.Image]:
    """

    Args:
        path:

    Returns:

    """
    try:
        return Image.open(path)
    except (OSError, AttributeError):
        return None


def is_image_truncated(image: Image) -> bool:
    """Returns True if the path refers to a truncated image

    Args:
        image:

    Returns:
        True if the image is truncated
    """
    try:
        image.load()
        return False
    except (OSError, AttributeError):
        return True


def get_image_shape(image: Image) -> Optional[Tuple[int, int]]:
    """

    Args:
        image:

    Returns:

    """
    try:
        return image.size
    except (OSError, AttributeError):
        return None


def hash_image(image: Image) -> Optional[str]:
    """

    Args:
        image:

    Returns:

    """
    try:
        return str(imagehash.phash(image))
    except (OSError, AttributeError):
        return None


def decode_byte_exif(exif_val: Union[str, bytes]) -> str:
    """Decode byte encodings

    Args:
        exif_val:

    Returns:

    """
    if isinstance(exif_val, str):
        return exif_val
    else:
        return exif_val.decode()


def extract_exif(image: Image) -> dict:
    """

    Args:
        image:

    Returns:

    """
    try:
        exif_data = image._getexif()
        if exif_data is not None:
            exif = {
                ExifTags.TAGS[k]: decode_byte_exif(v)
                for k, v in exif_data.items()
                if k in ExifTags.TAGS
            }
        else:
            exif = {}
    except (AttributeError, OSError):
        # Not all file types (e.g. .gif) have exif information.
        exif = {}

    return exif


def path_is_image(p: Path) -> bool:
    return imghdr.what(p) is not None



import numpy as np
import pandas as pd
import pytest

from visions.dtypes.boolean import BoolDtype


@pytest.mark.parametrize(
    "series,expected_values,expected_dtype",
    [
        (pd.Series([True, False], dtype="Bool"), (True, False), "Bool"),
        (pd.Series([True, False], dtype=bool), (True, False), bool),
        (
            pd.Series([True, False, None, True, False, None], dtype="Bool"),
            (True, False, None, True, False, None),
            "Bool",
        ),
        (
            pd.Series([True, False, None, True, False, None], dtype=bool),
            (True, False, False, True, False, False),
            bool,
        ),
        (
            pd.Series([True, False, np.nan, True, False, np.nan], dtype="Bool"),
            (True, False, None, True, False, None),
            "Bool",
        ),
        (
            pd.Series([True, False, np.nan, True, False, np.nan], dtype=bool),
            (True, False, True, True, False, True),
            bool,
        ),
    ],
)
def test_series(series, expected_values, expected_dtype):
    value_series = pd.Series(expected_values)
    assert series[series.notna()].eq(value_series[value_series.notna()]).all()
    assert series.dtype == expected_dtype



from ipaddress import _BaseAddress, ip_address
from typing import Sequence

import pandas as pd

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.coercion import test_utils
from visions.utils.series_utils import nullable_series_contains


def to_ip(series: pd.Series) -> pd.Series:
    return series.apply(ip_address)


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Object, String

    relations = [
        IdentityRelation(cls, Object),
        InferenceRelation(
            cls, String, relationship=test_utils.coercion_test(to_ip), transformer=to_ip
        ),
    ]
    return relations


class IPAddress(VisionsBaseType):
    """**IP Address** (v4 and v6) implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> from ipaddress import IPv4Address
        >>> x = pd.Series([IPv4Address('127.0.0.1'), IPv4Address('128.0.1.2')])
        >>> x in visions.IPAddress
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    @nullable_series_contains
    def contains_op(cls, series: pd.Series) -> bool:
        return all(isinstance(x, _BaseAddress) for x in series)



from pathlib import Path
from typing import Union

import networkx as nx


def output_graph(G: nx.DiGraph, file_name: Union[Path, str], sort: bool = True) -> None:
    """Output a graph to a file, either as image or as dot file.

    Args:
        G: the DiGraph to write or plot
        file_name: the file name to write to. Extension can be svg, png or dot.
        sort: create a copy of the graph with sorted keys

    Returns:
        Nothing

    Raises:
        ValueError when the file_name does not end on .svg, .png or .dot
    """

    if sort:
        # Create ordered graph for deterministic image outputs
        G_sorted = nx.DiGraph()
        G_sorted.graph["node"] = {"shape": "box", "color": "red"}
        G_sorted.add_nodes_from(sorted(G.nodes, key=lambda x: str(x)))
        G_sorted.add_edges_from(sorted(G.edges, key=lambda x: (str(x[0]), str(x[1]))))
        G = G_sorted

    p = nx.drawing.nx_pydot.to_pydot(G)
    if not isinstance(file_name, Path):
        file_name = Path(file_name)

    if file_name.suffix == ".svg":
        p.write_svg(file_name)
    elif file_name.suffix == ".png":
        p.write_png(file_name)
    elif file_name.suffix == ".dot":
        p.write_dot(file_name)
    else:
        raise ValueError("Extension should be .dot, .svg or .png")



from typing import Sequence

import numpy as np
import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.float import test_is_float
from visions.types.type import VisionsBaseType
from visions.utils.coercion import test_utils


def test_string_is_complex(series) -> bool:
    coerced_series = test_utils.option_coercion_evaluator(to_complex)(series)
    return coerced_series is not None and not test_is_float(series)


def to_complex(series: pd.Series) -> bool:
    return series.apply(complex)


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Generic, String

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(
            cls, String, relationship=test_string_is_complex, transformer=to_complex
        ),
    ]
    return relations


class Complex(VisionsBaseType):
    """**Complex** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> import numpy as np
        >>> x = pd.Series([np.complex(0, 0), np.complex(1, 2), np.complex(3, -1), np.nan])
        >>> x in visions.Complex
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return pdt.is_complex_dtype(series)



from typing import Sequence

import pandas as pd

from visions.relations import TypeRelation
from visions.types.type import VisionsBaseType


class Generic(VisionsBaseType):
    """**Generic** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> import numpy as np
        >>> x = pd.Series(['a', 1, np.nan])
        >>> x in visions.Generic
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return []

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return True



from visions.application.summaries.series.base_summary import base_summary
from visions.application.summaries.series.category_summary import category_summary
from visions.application.summaries.series.email_address_summary import (
    email_address_summary,
)
from visions.application.summaries.series.file_summary import file_summary
from visions.application.summaries.series.image_summary import image_summary
from visions.application.summaries.series.missing_summary import missing_summary
from visions.application.summaries.series.numerical_basic_summary import (
    numerical_basic_summary,
)
from visions.application.summaries.series.numerical_summary import numerical_summary
from visions.application.summaries.series.path_summary import path_summary
from visions.application.summaries.series.range_summary import range_summary
from visions.application.summaries.series.text_summary import text_summary
from visions.application.summaries.series.unique_summary import (
    unique_summary,
    unique_summary_complex,
)
from visions.application.summaries.series.url_summary import url_summary
from visions.application.summaries.series.zero_summary import zero_summary



import pandas as pd


def category_summary(series: pd.Series) -> dict:
    """Summary for pandas.Categorical

    Args:
        series: series to summarize

    Returns:
        A summary of the series including `category_size`, `ordered` and `missing_categorical_values`.
    """
    summary = {
        "category_size": len(series.dtype._categories),
        "ordered": series.dtype.ordered,
    }

    summary["missing_categorical_values"] = (
        True if series.nunique() != summary["category_size"] else False
    )
    return summary



import pandas as pd
import pytest

from tests.series import get_series
from tests.utils import (
    contains,
    convert,
    get_contains_cases,
    get_convert_cases,
    get_inference_cases,
    infers,
)
from visions import CompleteSet
from visions.types import (
    URL,
    UUID,
    Boolean,
    Categorical,
    Complex,
    Count,
    Date,
    DateTime,
    EmailAddress,
    File,
    Float,
    Generic,
    Geometry,
    Image,
    Integer,
    IPAddress,
    Object,
    Ordinal,
    Path,
    String,
    Time,
    TimeDelta,
)

series = get_series()

typeset = CompleteSet()

contains_map = {
    Integer: {
        "int_series",
        "Int64_int_series",
        "int_range",
        "Int64_int_nan_series",
        "int_series_boolean",
    },
    Count: {"np_uint32"},
    Path: {"path_series_linux", "path_series_linux_missing", "path_series_windows"},
    URL: {"url_series", "url_nan_series", "url_none_series"},
    Float: {
        "float_series",
        "float_series2",
        "float_series3",
        "float_series4",
        "inf_series",
        "nan_series",
        "float_nan_series",
        "float_series5",
        "int_nan_series",
        "nan_series_2",
        "float_with_inf",
        "float_series6",
    },
    Categorical: {
        "categorical_int_series",
        "categorical_float_series",
        "categorical_string_series",
        "categorical_complex_series",
        "categorical_char",
        "ordinal",
    },
    Boolean: {"bool_series", "bool_series2", "bool_series3", "nullable_bool_series"},
    Complex: {
        "complex_series",
        "complex_series_py",
        "complex_series_nan",
        "complex_series_py_nan",
        "complex_series_nan_2",
        "complex_series_float",
    },
    DateTime: {
        "timestamp_series",
        "timestamp_aware_series",
        "datetime",
        "timestamp_series_nat",
        "date_series_nat",
    },
    Date: {"date"},
    Time: {"time"},
    TimeDelta: {"timedelta_series", "timedelta_series_nat", "timedelta_negative"},
    String: {
        "timestamp_string_series",
        "string_with_sep_num_nan",
        "string_series",
        "geometry_string_series",
        "string_unicode_series",
        "string_np_unicode_series",
        "path_series_linux_str",
        "path_series_windows_str",
        "int_str_range",
        "string_date",
        "textual_float",
        "textual_float_nan",
        "ip_str",
        "string_flt",
        "string_num",
        "str_url",
        "string_str_nan",
        "string_num_nan",
        "string_bool_nan",
        "string_flt_nan",
        "str_complex",
        "uuid_series_str",
        "str_int_leading_zeros",
        "email_address_str",
        "str_float_non_leading_zeros",
        "str_int_zeros",
    },
    Geometry: {"geometry_series", "geometry_series_missing"},
    IPAddress: {"ip", "ip_mixed_v4andv6", "ip_missing"},
    Ordinal: {"ordinal"},
    UUID: {"uuid_series", "uuid_series_missing"},
    File: {
        "file_test_py",
        "file_mixed_ext",
        "file_test_py_missing",
        "image_png",
        "image_png_missing",
    },
    Image: {"image_png", "image_png_missing"},
    EmailAddress: {"email_address", "email_address_missing"},
}

if int(pd.__version__[0]) >= 1:
    contains_map[String].add("string_dtype_series")

contains_map[Object] = {
    "mixed_list[str,int]",
    "mixed_dict",
    "callable",
    "module",
    "mixed_integer",
    "mixed_list",
    "mixed",
    "bool_nan_series",
}

# Empty series
contains_map[Generic] = {
    "empty",
    "empty_bool",
    "empty_float",
    "empty_int64",
    "empty_object",
}


@pytest.mark.parametrize(**get_contains_cases(series, contains_map, typeset))
def test_contains(series, type, member):
    """Test the generated combinations for "series in type"

    Args:
        series: the series to test
        type: the type to test against
        member: the result
    """
    result, message = contains(series, type, member)
    assert result, message


inference_map = {
    "int_series": Integer,
    "categorical_int_series": Categorical,
    "int_nan_series": Integer,
    "Int64_int_series": Integer,
    "Int64_int_nan_series": Integer,
    "np_uint32": Count,
    "int_range": Integer,
    "float_series": Float,
    "float_nan_series": Float,
    "int_series_boolean": Integer,
    "float_series2": Integer,
    "float_series3": Float,
    "float_series4": Float,
    "float_series5": Float,
    "float_series6": Float,
    "complex_series_float": Integer,
    "categorical_float_series": Categorical,
    "float_with_inf": Float,
    "inf_series": Float,
    "nan_series": Float,
    "nan_series_2": Float,
    "string_series": String,
    "categorical_string_series": Categorical,
    "timestamp_string_series": Date,
    "string_with_sep_num_nan": String,  # TODO: Introduce thousands separator
    "string_unicode_series": String,
    "string_np_unicode_series": String,
    "string_num_nan": Integer,
    "string_num": Integer,
    "string_flt_nan": Float,
    "string_flt": Float,
    "string_str_nan": String,
    "string_bool_nan": Boolean,
    "int_str_range": Integer,
    "string_date": Date,
    "str_url": URL,
    "bool_series": Boolean,
    "bool_nan_series": Boolean,
    "nullable_bool_series": Boolean,
    "bool_series2": Boolean,
    "bool_series3": Boolean,
    "complex_series": Complex,
    "complex_series_nan": Complex,
    "complex_series_nan_2": Complex,
    "complex_series_py_nan": Complex,
    "complex_series_py": Complex,
    "categorical_complex_series": Categorical,
    "timestamp_series": DateTime,
    "timestamp_series_nat": DateTime,
    "timestamp_aware_series": DateTime,
    "datetime": Date,
    "timedelta_series": TimeDelta,
    "timedelta_series_nat": TimeDelta,
    "timedelta_negative": TimeDelta,
    "geometry_string_series": Geometry,
    "geometry_series_missing": Geometry,
    "geometry_series": Geometry,
    "path_series_linux": Path,
    "path_series_linux_missing": Path,
    "path_series_linux_str": Path,
    "path_series_windows": Path,
    "path_series_windows_str": Path,
    "url_series": URL,
    "url_nan_series": URL,
    "url_none_series": URL,
    "mixed_list[str,int]": Object,
    "mixed_dict": Object,
    "mixed_integer": Object,
    "mixed_list": Object,
    "mixed": Boolean,
    "callable": Object,
    "module": Object,
    "textual_float": Float,
    "textual_float_nan": Float,
    "empty": Generic,
    "empty_object": Generic,
    "empty_float": Generic,
    "empty_bool": Generic,
    "empty_int64": Generic,
    "ip": IPAddress,
    "ip_str": IPAddress,
    "ip_missing": IPAddress,
    "date_series_nat": Date,
    "date": Date,
    "time": Time,
    "categorical_char": Categorical,
    "ordinal": Ordinal,
    "str_complex": Complex,
    "uuid_series": UUID,
    "uuid_series_str": UUID,
    "uuid_series_missing": UUID,
    "ip_mixed_v4andv6": IPAddress,
    "file_test_py": File,
    "file_test_py_missing": File,
    "file_mixed_ext": File,
    "image_png": Image,
    "image_png_missing": Image,
    "str_int_leading_zeros": String,
    "str_float_non_leading_zeros": Float,
    "str_int_zeros": Integer,
    "email_address": EmailAddress,
    "email_address_missing": EmailAddress,
    "email_address_str": EmailAddress,
}
if int(pd.__version__[0]) >= 1:
    inference_map["string_dtype_series"] = String


@pytest.mark.parametrize(**get_inference_cases(series, inference_map, typeset))
def test_inference(series, type, typeset, difference):
    """Test the generated combinations for "inference(series) == type"

    Args:
        series: the series to test
        type: the type to test against
    """
    result, message = infers(series, type, typeset, difference)
    assert result, message


# Conversions in one single step
convert_map = [
    # Model type, Relation type
    (Integer, Float, {"int_nan_series", "float_series2"}),
    (Complex, String, {"str_complex"}),
    (
        Float,
        String,
        {
            "string_flt",
            "string_num_nan",
            "string_num",
            "string_flt_nan",
            "textual_float",
            "textual_float_nan",
            "int_str_range",
            "str_float_non_leading_zeros",
            "str_int_zeros",
            # "string_with_sep_num_nan",
        },
    ),
    (Date, DateTime, {"date_series_nat", "datetime"}),
    (DateTime, String, {"timestamp_string_series", "string_date"}),
    (Geometry, String, {"geometry_string_series"}),
    (Boolean, String, {"string_bool_nan"}),
    (IPAddress, String, {"ip_str"}),
    (URL, String, {"str_url"}),
    (Path, String, {"path_series_windows_str", "path_series_linux_str"}),
    (EmailAddress, String, {"email_address_str"}),
    (Float, Complex, {"complex_series_float"}),
    (Boolean, Object, {"bool_nan_series", "mixed"}),
    (UUID, String, {"uuid_series_str"}),
]


@pytest.mark.parametrize(**get_convert_cases(series, convert_map, typeset))
def test_conversion(source_type, relation_type, series, member):
    """Test the generated combinations for "convert(series) == type" and "infer(series) = source_type"

    Args:
        series: the series to test
        type: the type to test against
    """
    result, message = convert(source_type, relation_type, series, member)
    assert result, message



from datetime import date, time
from typing import Sequence

import pandas as pd

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.coercion import test_utils
from visions.utils.series_utils import class_name_attrs, nullable_series_contains


def test_date(series):
    dtseries = series.dropna().dt.time
    value = time(0, 0)
    return True if all(v == value for v in dtseries) else None


def to_date(series):
    return series.dt.date


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import DateTime, Object

    relations = [
        IdentityRelation(cls, Object),
        InferenceRelation(
            cls,
            DateTime,
            relationship=test_utils.coercion_test(test_date),
            transformer=to_date,
        ),
    ]
    return relations


class Date(VisionsBaseType):
    """**Date** implementation of :class:`visions.types.type.VisionsBaseType`.
    All values are should be datetime.date or missing

    Examples:
        >>> x = pd.Series([datetime.date(2017, 3, 5), datetime.date(2019, 12, 4)])
        >>> x in visions.Date
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    @nullable_series_contains
    def contains_op(cls, series: pd.Series) -> bool:
        return class_name_attrs(series, date, ["year", "month", "day"])



import os

import pandas as pd


def path_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    summary = {
        "common_prefix": os.path.commonprefix(list(series)) or "No common prefix",
        "stem_counts": series.map(lambda x: x.stem).value_counts(),
        "suffix_counts": series.map(lambda x: x.suffix).value_counts(),
        "name_counts": series.map(lambda x: x.name).value_counts(),
        "parent_counts": series.map(lambda x: x.parent).value_counts(),
        "anchor_counts": series.map(lambda x: x.anchor).value_counts(),
    }

    summary["n_stem_unique"] = len(summary["stem_counts"])
    summary["n_suffix_unique"] = len(summary["suffix_counts"])
    summary["n_name_unique"] = len(summary["name_counts"])
    summary["n_parent_unique"] = len(summary["parent_counts"])
    summary["n_anchor_unique"] = len(summary["anchor_counts"])

    return summary



import pandas as pd


def email_address_summary(series: pd.Series) -> dict:
    summary = {}

    local = []
    fqdn = []
    for v in series.drop().values:
        local.append(v.local)
        fqdn.append(v.fqdn)

    summary["local_counts"] = pd.Series(local).value_counts().to_dict()
    summary["fqdn_counts"] = pd.Series(fqdn).value_counts().to_dict()

    return summary



from pathlib import Path

from visions.typesets import CompleteSet, GeometrySet, StandardSet

# Windows Note
# Tip for Python3/64-bit compatible version of pygraphviz
# https://github.com/CristiFati/Prebuilt-Binaries/raw/master/Windows/PyGraphviz/pygraphviz-1.5-cp37-cp37m-win_amd64.whl


def generate_typeset_plots():
    typesets_dir = Path("typesets/")
    typesets_dir.mkdir(exist_ok=True)

    # Initialize typeset
    for name, tsc in [
        ("typeset_complete", CompleteSet()),
        ("typeset_geometry", GeometrySet()),
        ("typeset_standard", StandardSet()),
    ]:
        # Write graph to dot
        tsc.output_graph(typesets_dir / "{name}.dot".format(name=name))

        # Plot the graph (svg)
        tsc.output_graph(typesets_dir / "{name}.svg".format(name=name))
        tsc.output_graph(
            typesets_dir / "{name}_base.svg".format(name=name), base_only=True
        )

        # Plot the graph (png)
        tsc.output_graph(typesets_dir / "{name}.png".format(name=name), dpi=150)


if __name__ == "__main__":
    generate_typeset_plots()



from typing import List, Sequence

import numpy as np
import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.coercion import test_utils


def to_int(series: pd.Series) -> pd.Series:
    try:
        return series.astype(int)
    except ValueError:
        return series.astype("Int64")


def float_is_int(series: pd.Series) -> bool:
    def check_equality(series):
        if series.empty or not np.isfinite(series).all():
            return False
        return series.eq(series.astype(int)).all()

    return check_equality(series.dropna() if series.hasnans else series)


def test_string_is_int(series) -> bool:
    coerced_series = test_utils.option_coercion_evaluator(string_to_int)(series)
    return coerced_series is not None and coerced_series in Integer


def string_to_int(series: pd.Series) -> pd.Series:
    # if any("," in x for x in series.dropna()):
    #     series = series.str.replace(",", "")

    return to_int(series)


def _get_relations(cls) -> List[TypeRelation]:
    from visions.types import Float, Generic

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(cls, Float, relationship=float_is_int, transformer=to_int),
    ]
    return relations


class Integer(VisionsBaseType):
    """**Integer** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([1, 2, 3])
        >>> x in visions.Integer
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return pdt.is_signed_integer_dtype(series)



from visions.types import (
    URL,
    UUID,
    Boolean,
    Categorical,
    Complex,
    Count,
    Date,
    DateTime,
    EmailAddress,
    File,
    Float,
    Generic,
    Geometry,
    Image,
    Integer,
    IPAddress,
    Object,
    Ordinal,
    Path,
    String,
    Time,
    TimeDelta,
)
from visions.typesets.typeset import VisionsTypeset


class CompleteSet(VisionsTypeset):
    """Complete visions typeset with all supported types

    Includes support for the following types:

    - Float
    - Integer
    - Boolean
    - Object
    - String
    - Complex
    - Categorical
    - Ordinal
    - Count
    - DateTime
    - Date
    - Time
    - TimeDelta
    - Geometry
    - Path
    - File
    - Image
    - URL
    - IPAddress
    - EmailAddress
    - UUID

    """

    def __init__(self):
        types = {
            Generic,
            Boolean,
            Float,
            Object,
            Complex,
            Categorical,
            Ordinal,
            DateTime,
            TimeDelta,
            Integer,
            Count,
            String,
            Geometry,
            URL,
            Path,
            Date,
            Time,
            File,
            Image,
            IPAddress,
            EmailAddress,
            UUID,
        }
        super().__init__(types)

        try:
            import imagehash
            import PIL
            import shapely
        except ImportError as e:
            raise ImportError(
                "This typeset requires dependencies that are currently not installed ({}). "
                "You can follow the installation instructions to resolve this issue: "
                "https://dylan-profiler.github.io/visions/visions/getting_started/installation.html"
                "".format(str(e))
            )



import pandas as pd

from visions.typesets import CompleteSet


def test_cast_copy():
    s = pd.Series(["1", "2", "3", "4"])
    id_s = hex(id(s))

    typeset = CompleteSet()
    ns = typeset.cast_to_inferred(s)
    id_ns = hex(id(ns))
    assert id_s != id_ns


def test_noncast_noncopy():
    s = pd.Series(["asdasd", "asdasda", "asdasd", "sadasd"])
    id_s = hex(id(s))

    typeset = CompleteSet()
    ns = typeset.cast_to_detected(s)
    id_ns = hex(id(ns))
    assert id_s == id_ns



from typing import Sequence

import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, TypeRelation
from visions.types.type import VisionsBaseType


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Object

    relations = [IdentityRelation(cls, Object)]

    return relations


pandas_has_string_dtype_flag = hasattr(pdt, "is_string_dtype")


class String(VisionsBaseType):
    """**String** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series(['rubin', 'carter', 'champion'])
        >>> x in visions.String
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        # TODO: without the object check this passes string categories... is there a better way?
        if pdt.is_categorical_dtype(series):
            return False
        elif not pdt.is_object_dtype(series):
            if pandas_has_string_dtype_flag and pdt.is_string_dtype(series):
                return True
            return False

        if series.hasnans:
            series = series.dropna()
            if series.empty:
                return False

        return all(isinstance(v, str) for v in series)



from typing import Dict, List


def get_boolean_coercions(id: str) -> List[Dict]:
    coercion_map = {
        "default": [{"true": True, "false": False}],
        "en": [
            {"true": True, "false": False},
            {"y": True, "n": False},
            {"yes": True, "no": False},
        ],
        "nl": [
            {"true": True, "false": False},
            {"ja": True, "nee": False},
            {"j": True, "n": False},
        ],
    }
    return coercion_map[id]


def get_language_bool(language_code: str):
    from visions.types.boolean import Boolean

    return Boolean.make_string_coercion(
        language_code, get_boolean_coercions(language_code)
    )



""" Utilities suite for visions """
from visions.utils.coercion import test_utils

# from visions.utils.images import image_utils
from visions.utils.monkeypatches import imghdr_patch, pathlib_patch
from visions.utils.profiling import profile_type
from visions.utils.series_utils import (
    func_nullable_series_contains,
    isinstance_attrs,
    nullable_series_contains,
)
from visions.utils.warning_handling import suppress_warnings



import functools
import timeit

import numpy as np
import pandas as pd


def big_o_tester(test_func):
    import big_o

    @functools.wraps(func)
    def inner(test_series):
        try:
            best, _ = big_o.big_o(
                test_func, lambda n: test_series[0:n], max_n=test_series.shape[0]
            )
            return best
        except np.linalg.LinAlgError:
            return np.nan

    return inner


def profile_type(dtype, profile_data, run_count=10, normed_length=100000):
    profile_data = {
        name: pd.Series(np.random.choice(data, normed_length))
        for name, data in profile_data.items()
        if len(data) > 0
    }
    big_O_test = big_o_tester(lambda x: x in dtype)
    return [
        {
            "type": dtype,
            "series": name,
            "run count": run_count,
            "average run time": timeit.timeit(lambda: data in dtype, number=run_count)
            / run_count,
            "big O": big_O_test(data),
        }
        for name, data in profile_data.items()
    ]


def profile_relation_is_relation(
    relation, profile_data, run_count=10, normed_length=100000
):
    profile_data = {
        name: pd.Series(np.random.choice(data, normed_length))
        for name, data in profile_data.items()
        if len(data) > 0
    }
    big_O_test = big_o_tester(relation.is_relation)
    return [
        {
            "relation": relation,
            "series": name,
            "run count": run_count,
            "average run time": timeit.timeit(
                lambda: relation.is_relation, number=run_count
            )
            / run_count,
            "big O": big_O_test(data),
        }
        for name, data in profile_data.items()
    ]


def profile_relation_transform(
    relation, profile_data, run_count=10, normed_length=100000
):
    profile_data = {
        name: pd.Series(np.random.choice(data, normed_length))
        for name, data in profile_data.items()
        if len(data) > 0
    }
    big_O_test = big_o_tester(relation.transform)
    return [
        {
            "relation": relation,
            "series": name,
            "run count": run_count,
            "average run time": timeit.timeit(
                lambda: relation.transform, number=run_count
            )
            / run_count,
            "big O": big_O_test(data),
        }
        for name, data in profile_data.items()
    ]



from visions.types import (
    Boolean,
    Categorical,
    Complex,
    DateTime,
    Float,
    Generic,
    Geometry,
    Integer,
    Object,
    String,
    TimeDelta,
)
from visions.typesets.typeset import VisionsTypeset


class GeometrySet(VisionsTypeset):
    """Standard visions typeset with shapely geometry support

    Includes support for the following types:

    - Float
    - Integer
    - Boolean
    - Object
    - String
    - Complex
    - Categorical
    - DateTime
    - TimeDelta
    - Geometry

    """

    def __init__(self):
        types = {
            Generic,
            Boolean,
            Float,
            Object,
            Complex,
            Categorical,
            DateTime,
            TimeDelta,
            Integer,
            String,
            Geometry,
        }
        super().__init__(types)

        try:
            import shapely
        except ImportError as e:
            raise ImportError(
                "This typeset requires dependencies that are currently not installed ({}). "
                "You can follow the installation instructions to resolve this issue: "
                "https://dylan-profiler.github.io/visions/visions/getting_started/installation.html"
                "".format(str(e))
            )



import pandas as pd

from visions import String
from visions.relations import InferenceRelation
from visions.utils.coercion import test_utils


def to_datetime_year_week(series: pd.Series) -> pd.Series:
    """Convert a series of the format YYYY/UU (year, week) to datetime.
    A '0' is added as day dummy value, as pandas requires a day value to parse.

    Args:
        series: the Series to parse

    Returns:
        A datetime series

    Examples:
        >>> series = pd.Series(['2018/47', '2018/12', '2018/03'])
        >>> parsed_series = to_datetime_year_week(series)
        >>> print(parsed_series.dt.week)
        0    47
        1    12
        2     3
        dtype: int64
    """
    return pd.to_datetime(series + "0", format="%Y/%U%w")


def to_datetime_year_month_day(series: pd.Series) -> pd.Series:
    """Convert a series of the format YYYYMMDD (year, month, day) to datetime.

    Args:
        series: the Series to parse

    Returns:
        A datetime series

    Examples:
        >>> series = pd.Series(['20181201', '20181202', '20181203'])
        >>> parsed_series = to_datetime_year_week(series)
        >>> print(parsed_series.dt.day)
        0    1
        1    2
        2    3
        dtype: int64
    """
    return pd.to_datetime(series, format="%Y%m%d")


def _to_datetime(cls, func) -> InferenceRelation:
    return InferenceRelation(
        relationship=test_utils.coercion_test(func),
        transformer=func,
        related_type=String,
        type=cls,
    )


def string_to_datetime_year_week(cls) -> InferenceRelation:
    return _to_datetime(cls, to_datetime_year_week)


def string_to_datetime_year_month_day(cls) -> InferenceRelation:
    return _to_datetime(cls, to_datetime_year_month_day)



from typing import List, Sequence

import numpy as np
import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types import Integer, Object, String, VisionsBaseType
from visions.utils.coercion.test_utils import coercion_map, coercion_map_test


def to_category(series: pd.Series) -> pd.Series:
    if series.isin({True, False}).all():
        return series.astype(bool)
    elif series.isin({True, False, None, np.nan}).all():
        return series.astype("Bool")
    else:
        unsupported_values = series[~series.isin({True, False, None, np.nan})].unique()
        raise ValueError(
            "Values not supported {unsupported_values}".format(
                unsupported_values=unsupported_values
            )
        )


def _get_relations(cls) -> List[TypeRelation]:
    from visions.types import Generic

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(
            cls,
            String,
            relationship=lambda s: coercion_map_test(cls.string_coercions)(
                s.str.lower()
            ),
            transformer=lambda s: to_category(
                coercion_map(cls.string_coercions)(s.str.lower())
            ),
        ),
        InferenceRelation(
            cls,
            Integer,
            relationship=lambda s: s.isin({0, 1, np.nan}).all(),
            transformer=to_category,
        ),
        InferenceRelation(
            cls,
            Object,
            relationship=lambda s: s.apply(type).isin([type(None), bool]).all(),
            transformer=to_category,
        ),
    ]
    return relations


class Category(VisionsBaseType):
    """**Categorical** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([True, False, 1], dtype='category')
        >>> x in visions.Category
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return pdt.is_categorical_dtype(series) or pdt.is_bool_dtype(series)



import pandas as pd

from visions.relations.relations import InferenceRelation
from visions.types.ordinal import to_ordinal


def check_consecutive(l: list) -> bool:
    return sorted(l) == list(range(min(l), max(l) + 1))


def is_ordinal_str(s: pd.Series) -> bool:
    if s.str.len().max() == 1:
        unique_values = list(s[s.notna()].str.lower().unique())
        return "a" in unique_values and check_consecutive(list(map(ord, unique_values)))
    else:
        return False


def string_to_ordinal(cls) -> InferenceRelation:
    from visions.types import String

    return InferenceRelation(
        type=cls,
        related_type=String,
        relationship=is_ordinal_str,
        transformer=to_ordinal,
    )



import numpy as np
import pandas as pd

from visions.relations.relations import InferenceRelation
from visions.types import Integer


def is_unsigned_int(series: pd.Series) -> bool:
    # TODO: add coercion, ensure that > uint.MAX raises error
    return series.ge(0).all()


def integer_to_count(cls) -> InferenceRelation:
    return InferenceRelation(
        relationship=is_unsigned_int,
        transformer=lambda s: s.astype(np.uint64),
        type=cls,
        related_type=Integer,
    )



import sys

import pytest

from tests.series import get_series
from visions.typesets import CompleteSet


def all_relations_tested(series_map):
    typeset = CompleteSet()

    # Convert data structure for mapping
    series_map_lookup = {}
    for map_to_type, map_from_type, items in series_map:
        try:
            series_map_lookup[map_to_type][map_from_type] = items
        except KeyError:
            series_map_lookup[map_to_type] = {map_from_type: items}

    missing_relations = set()
    for node in typeset.types:
        for relation in node.relations:
            from_type, to_type = relation.related_type, relation.type
            if relation.inferential and (
                to_type not in series_map_lookup
                or from_type not in series_map_lookup[to_type]
                or len(series_map_lookup[to_type][from_type]) == 0
            ):
                missing_relations.add(str(relation))

    if len(missing_relations) > 0:
        raise ValueError(
            "Not all inferential relations are tested {missing_relations}".format(
                missing_relations=missing_relations
            )
        )


def pytest_generate_tests(metafunc):
    _test_suite = get_series()
    if metafunc.function.__name__ in [
        "test_consistency",
        "test_side_effects",
        "test_multiple_inference",
    ]:
        argsvalues = []
        for series in _test_suite:
            args = {"id": str(series.name)}
            argsvalues.append(pytest.param(series, **args))

        metafunc.parametrize(argnames=["series"], argvalues=argsvalues)


def test_consistency(series):
    typeset = CompleteSet()

    if (
        series.name
        in ["timedelta_series_nat", "date_series_nat", "timestamp_series_nat"]
        and sys.version_info.major == 3
        and sys.version_info.minor == 7
    ):
        pytest.skip("unsupported configuration")

    initial_type = typeset.detect_type(series.copy(deep=True))
    converted_type = typeset.infer_type(series.copy(deep=True))

    if initial_type != converted_type:
        converted_series = typeset.cast_to_inferred(series.copy(deep=True))
        assert series.dtype.kind != converted_series.dtype.kind or not (
            (
                converted_series.eq(series) | (converted_series.isna() & series.isna())
            ).all()
        )
    else:
        converted_series = typeset.cast_to_inferred(series.copy(deep=True))
        assert (
            converted_series.eq(series) | (converted_series.isna() & series.isna())
        ).all()


def test_side_effects(series):
    reference = series.copy()

    typeset = CompleteSet()
    typeset.detect_type(series)
    typeset.infer_type(series)

    # Check if NaN mask is equal
    assert series.notna().eq(reference.notna()).all()
    # Check if NonNaN values are equal
    assert series[series.notna()].eq(reference[reference.notna()]).all()


def test_multiple_inference(series):
    """
    Notes:
        Copy to prevent possible side effects only for testing.
    """
    ts = CompleteSet()

    inferred_type = ts.infer_type(series)

    series_convert = ts.cast_to_inferred(series.copy(deep=True))

    initial_type_after_convert = ts.detect_type(series_convert.copy(deep=True))
    assert inferred_type == initial_type_after_convert

    series_convert2 = ts.cast_to_inferred(series_convert.copy(deep=True))

    inferred_type_after_convert = ts.infer_type(series_convert2.copy(deep=True))
    assert initial_type_after_convert == inferred_type_after_convert

    assert series_convert.isna().eq(series_convert2.isna()).all()
    assert (
        series_convert[series_convert.notna()]
        .eq(series_convert2[series_convert2.notna()])
        .all()
    )



import json
import re
from pathlib import Path

import networkx as nx

from visions.typesets import CompleteSet


def update(data):
    data["name"] = data.pop("id")
    if "children" not in data:
        data["size"] = 1
    else:
        data["children"] = [update(child) for child in data["children"]]
    return data


def write_json(data):
    with Path("typesets/typeset_complete_base.json").open("w") as f:
        json.dump(data, f)


def write_html(data):
    string = "\n\troot = {jdata};\n\t".format(jdata=json.dumps(data))

    file_name = Path("circular_packing.html")
    fc = file_name.read_text()
    fc = re.sub(
        r"// START-REPLACE(.*)// END-REPLACE",
        r"// START-REPLACE{string}// END-REPLACE".format(string=string),
        fc,
        flags=re.MULTILINE | re.DOTALL,
    )
    file_name.write_text(fc)


def to_json_tree_sorted(G, root):
    # json_graph.tree_data with sorting
    from itertools import chain

    def add_children(n, G):
        nbrs = G[n]
        if len(nbrs) == 0:
            return []
        children_ = []
        for child in nbrs:
            d = dict(chain(G.nodes[child].items(), [("id", child)]))
            c = add_children(child, G)
            if c:
                d["children"] = c
            children_.append(d)

        children_ = sorted(children_, key=lambda x: x["id"])
        return children_

    data = dict(chain(G.nodes[root].items(), [("id", root)]))
    data["children"] = add_children(root, G)
    return data


def write_circular_packing_files():
    typeset = CompleteSet()
    graph = typeset.base_graph.copy()
    nx.relabel_nodes(graph, {n: str(n) for n in graph.nodes}, copy=False)

    data = to_json_tree_sorted(graph, root="Generic")
    data = update(data)

    write_json(data)
    write_html(data)


if __name__ == "__main__":
    write_circular_packing_files()



from typing import Sequence

import numpy as np
import pandas as pd
from pandas.api import types as pdt

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.type import VisionsBaseType
from visions.utils.coercion import test_utils
from visions.utils.series_utils import func_nullable_series_contains
from visions.utils.warning_handling import suppress_warnings


def test_string_leading_zeros(series: pd.Series, coerced_series: pd.Series):
    if coerced_series.hasnans:
        notna = coerced_series.notna()
        coerced_series = coerced_series[notna]

        if coerced_series.empty:
            return False
        series = series[notna]

    return not any(s[0] == "0" for s in series[coerced_series > 1])


def test_string_is_float(series: pd.Series) -> bool:
    coerced_series = test_utils.option_coercion_evaluator(string_to_float)(series)
    if coerced_series is not None and coerced_series in Float:
        return test_string_leading_zeros(series, coerced_series)

    return False


def string_to_float(series: pd.Series) -> pd.Series:
    # Slightly faster to check for the character if it's not present than to
    # attempt the replacement
    # if any("," in x for x in series.dropna()):
    #     series = series.str.replace(",", "")

    return to_float(series)


def test_is_float(series: pd.Series) -> bool:
    coerced_series = test_utils.option_coercion_evaluator(to_float)(series)
    return coerced_series is not None and coerced_series in Float


def to_float(series: pd.Series) -> pd.Series:
    return series.astype(float)


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Complex, Generic, String

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(
            cls, String, relationship=test_string_is_float, transformer=string_to_float
        ),
        InferenceRelation(
            cls,
            Complex,
            relationship=lambda s: all(np.imag(s.values) == 0),
            transformer=suppress_warnings(to_float),
        ),
    ]
    return relations


class Float(VisionsBaseType):
    """**Float** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([1.0, 2.5, 5.0, np.nan])
        >>> x in visions.Float
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return pdt.is_float_dtype(series)
